import numpy as np
import os
from PIL import Image
import exifread
from matplotlib import pyplot as plt
from scipy.stats import linregress
from scipy.optimize import curve_fit

def get_tiff(search_path):
	dataset = {}	
	for folder in os.listdir(search_path):
		temp = float(folder)
		for filename in os.listdir(os.path.join(search_path,folder)):
			cr2_path = os.path.join(search_path,folder,filename)
			if os.path.splitext(cr2_path)[1] == '.CR2':
				tiff_path = os.path.splitext(cr2_path)[0] + '.tiff'
				if not os.path.exists(tiff_path):
					print 'Converting %s to tiff'%(filename)
					#os.system('dcraw.exe -d -t 0 -r 1 1 1 1 -4 -T %s'%cr2_path)
					os.system('dcraw.exe -t 0 -D -4 -T %s'%cr2_path)
				
				f = open(tiff_path,'rb')
				exif = exifread.process_file(f,details=False)
				f.close()

				imageinfo = {}				
				iso = exif['EXIF ISOSpeedRatings'].values[0]
				numerator = exif['EXIF ExposureTime'].values[0].num
				denumerator = exif['EXIF ExposureTime'].values[0].den
				exposure = float(numerator)/float(denumerator)
				imageinfo['Filename'] = tiff_path
				imageinfo['Temperature'] = temp
				imageinfo['Exposure'] = exposure
				imageinfo['ISO'] = iso
				imageinfo['Camera'] = '%s %s'%(exif['Image Make'],exif['Image Model'])
				imageinfo['BitsPerPixel'] = exif['Image BitsPerSample'].values[0]
				
				if not temp in dataset:
					dataset[temp] = {}
				if not iso in dataset[temp]:
					dataset[temp][iso] = {}
				if not exposure in dataset[temp][iso]:
					dataset[temp][iso][exposure] = []
					
				dataset[temp][iso][exposure].append(imageinfo)
				
	return dataset

def ro_noise_fit(gain, before, after):
	return np.sqrt(after**2 + (gain*before)**2)
	
def crop(data,size):
	x,y = data.shape
	x_a = int(round(((x-size)/2)/2)*2)
	x_b = x_a + size
	y_a = int(round(((y-size)/2)/2)*2)
	y_b = y_a + size
	
	return data[x_a:x_b,y_a:y_b].flatten().astype(np.double)
	
iso_bias = {}
	
def calc_gain(foldername):
	
	dataset = get_tiff(foldername)

	square_size = 512
	temp = dataset.keys()[0]
	iso_values = sorted(dataset[temp].keys())
	iso_gain = []
	iso_ronoise = []
	
	for iso in iso_values:
		# sort images by exposure
		bias_exp = np.min(dataset[temp][iso].keys())
		flat_exp = np.max(dataset[temp][iso].keys())

		# load images and extract pixel data to numpy array
		biasA_data = np.array(Image.open(dataset[temp][iso][bias_exp][0]['Filename']))
		biasB_data = np.array(Image.open(dataset[temp][iso][bias_exp][1]['Filename']))
		flatA_data = np.array(Image.open(dataset[temp][iso][flat_exp][0]['Filename']))
		flatB_data = np.array(Image.open(dataset[temp][iso][flat_exp][1]['Filename']))
		
		# crop and convert to 64-bit int
		biasA_sub = crop(biasA_data,square_size).astype(np.int64)
		biasB_sub = crop(biasB_data,square_size).astype(np.int64)
		flatA_sub = crop(flatA_data,square_size).astype(np.int64)
		flatB_sub = crop(flatB_data,square_size).astype(np.int64)

		# calculate mean for bias and flats
		mean_bias = (np.mean(biasA_sub) + np.mean(biasB_sub))/2
		mean_flat = (np.mean(flatA_sub) + np.mean(flatB_sub))/2
		mean = mean_flat - mean_bias
		
		# calculate variance for bias and flats
		var_bias = np.var(biasA_sub - biasB_sub)/2
		var_flat = np.var(flatA_sub - flatB_sub)/2
		var = var_flat - var_bias
		
		iso_bias[iso] = np.sqrt(var_bias)
		
		# calculate gain and readout noise
		gain = mean/var
		shot_noise = np.sqrt(var_flat)
		ro_noise = np.sqrt(var_bias)
		
		iso_gain.append(1./gain)
		iso_ronoise.append(ro_noise)
		
		# print and plot results
		print 'iso%4d - gain %3.2f APU/e- - noise %f'%(iso,1./gain,shot_noise)
		for ax in [ax_save,ax_gain]:
			ax.plot([mean_bias, mean_flat],[var_bias, var_flat],'o-',label='ISO%04d %3.2f APU/e-'%(iso,1./gain))
			
	
	#calculate unity gain
	slope, intercept, r,  p, std = linregress(iso_values,iso_gain)
	unity_gain = (1.-intercept)/slope
	print 'Unity gain ISO%d , r^2 %3.2f'%(unity_gain,r)
	
	# plot results
	for ax in [ax_save,ax_gain]:
		ax.legend(loc=2,prop={'size':10})
		ax.set_ylabel(u'Noise Var (APU\u00B2)')
		ax.set_xlabel('Signal (APU)')
		ax.grid(which='both')
		ax.set_title('Variance vs Signal (gain)')
	fig_save.tight_layout()
	fig_save.savefig('gain.png')
	ax_save.cla()
	

	iso_gain = np.array(iso_gain)
	iso_ronoise = np.array(iso_ronoise)

	gain_range = np.linspace(iso_gain[0],iso_gain[-1],100, True)
	p_opt, p_cov = curve_fit(ro_noise_fit,iso_gain, iso_ronoise)
	noise_after = p_opt[1]
	noise_before = p_opt[0]

	print 'Nb %3.1f - Na %3.1f'%(noise_before,noise_after)
	
	print iso_gain
	print gain_range
	print iso_ronoise
	print ro_noise_fit(gain_range,noise_before,noise_after)
	ax_save.plot(iso_gain,iso_ronoise,'bo')
	ax_save.plot(gain_range,ro_noise_fit(gain_range,noise_before,noise_after),'r:')
	fig_save.tight_layout()
	fig_save.savefig('readout_noise.png')
	ax_save.cla()
	
	# plot unity gain
	for ax in [ax_save,ax_unity]:
		ax.plot(iso_values,iso_gain,'o-',label='ISO%d-%d'%(iso_values[0],iso_values[-1]))
		ax.axvline(unity_gain, ls='--', color='purple', label='Unity gain ISO%3d'%unity_gain)
		ax.axhline(1., ls='--', color='purple')
		ax.legend(loc=0,prop={'size':10})
		ax.grid()
		ax.set_yscale('log')
		ax.set_yticks(iso_gain)
		ax.yaxis.set_ticklabels( ['%3.2f' % iso for iso in iso_gain] )
		ax.set_xscale('log')
		ax.set_xticks(iso_values)
		ax.xaxis.set_ticklabels( ['%d' % iso for iso in iso_values] )
		ax.set_ylabel('Gain apu/e- (log)')
		ax.set_xlabel('ISO (log)')
		ax.set_title('Gain vs ISO')
	fig_save.tight_layout()
	fig_save.savefig('unity.png')
	ax_save.cla()


def calc_noise(foldername):
	dataset = get_tiff(foldername)

	square_size = 512
	temp_values = sorted(dataset.keys())
	iso_values = None
	exp_values = None
	
	for temp,i in zip(temp_values,range(len(temp_values))):
		if not iso_values:
			iso_values = sorted(dataset[temp].keys())
		for iso in iso_values:
			if not exp_values:
				exp_values = sorted(dataset[temp][iso].keys())
			
			for exp in exp_values:
				image = dataset[temp][iso][exp][0]
				data = np.array(Image.open(image['Filename']))
			
				# crop and convert to 64-bit int
				sub = crop(data,square_size).astype(np.double)

				# calculate mean and std
				mean = np.mean(sub)
				std = np.std(sub)

				image['mean'] = mean
				image['noise'] = std

	for exp,i in zip(exp_values,range(7)):
		for iso in iso_values:
			noise = []
			for temp in temp_values:
				img = dataset[temp][iso][exp][0]
				noise_value = img['noise']
				mean_value = img['mean']
				noise.append(noise_value)
				dark_current = np.sqrt(noise_value**2 - iso_bias[iso]**2)/exp
				print u'%3d\u00B0CC  - iso%04d - exp %ds - noise %3.1f APU - %f mean'%(temp,iso,exp,noise_value,mean_value)
			
			ax_save.plot(temp_values,noise,'x-',label='ISO%d'%iso)
			ax_save.legend(loc=2,prop={'size':10})
			ax_save.set_ylabel('Noise std (APU)')
			ax_save.set_xlabel(u'Temperature (\u00B0C)')
			ax_save.grid(which='both')
			ax_save.set_title('Noise vs temperature @ %ds exposure'%exp);
			#ax.axhline(iso_bias[iso],ls=':')
		
		ax_save.set_ylim(0,70)

		fig_save.tight_layout()
		fig_save.savefig('noise_vs_temp_%ds.png'%exp)
		ax_save.cla()

	for iso in iso_values:
		for temp in temp_values:
			noise = []
			for exp in exp_values:
				noise_value = dataset[temp][iso][exp][0]['noise']
				noise.append(noise_value)
			
			ax_save.plot(exp_values,noise,'x-',label=u'%d\u00B0C'%temp)
		ax_save.legend(loc=2,prop={'size':10})
		ax_save.set_ylabel('Noise std (APU)')
		ax_save.set_xlabel('Exposure (s)')
		ax_save.grid(which='both')
		ax_save.set_title('Noise vs exposure @ ISO%d'%iso);
		ax_save.set_ylim(0,70)
		fig_save.tight_layout()
		fig_save.savefig('noise_vs_exp_iso%d.png'%iso)
		ax_save.cla()	
		
	# plot unity gain
	
fig_save = plt.figure(figsize=(6, 4))
ax_save = fig_save.add_subplot(111)

fig = plt.figure()
ax_gain = fig.add_subplot(221)
ax_unity = fig.add_subplot(222)
ax_ro_noise = fig.add_subplot(223)

	
calc_gain('e:/temp/test/gain')
calc_noise('e:/temp/test/dark')
fig.tight_layout()

plt.show()
